/*----------------------------------------------------------------------
  File     : AprioriSets.h
  Contents : apriori algorithm for finding frequent sets
  Author   : Bart Goethals
  Update   : 1/6/2003 
----------------------------------------------------------------------*/

#include <set>
#include <map>
#include <vector>
#include <fstream>
#include <string>
using namespace std;
#include "Data.h"
#include "Item.h"

/*NUOVO*/
#define MAX_LINE_LENGTH 2000
#define IP_NAME_FILE "certificate/ip.txt"
#include "ItemSet.h"
/*FINE NUOVO*/

class Element
{
public:
	Element(int iold, int inew=0) : oldid(iold), id(inew){}

	int oldid;
	int id;
	bool operator< (const Element  &e) const {return oldid < e.oldid;}
};

class AprioriSets 
{
public:

	AprioriSets();
	~AprioriSets();

	void setData(char *fn, int type);
	int setOutputSets(char *fn);
	void setMinSup(int ms){minsup=ms;}
	int generateSets();

	void setVerbose(){verbose=true;}
	void setCountType(int t){countType=t;}

	/*NEW*/
	struct ltstr
	{
	     bool operator() (char* const s1, char* const s2) const 
		  {
		       return strcmp(s1, s2) < 0;
		  }
	};

	unsigned long minsup;//supporto globale

	unsigned long minsuploc;//supporto locale

	unsigned long ntraloc;//numero transazioni locali

	unsigned long ntraglo;//numero transazioni globali

	FILE *fprint;//file su cui scrivere i valori;

	map <char*, ItemSet*, ltstr> myElements;

	unsigned long nItemSet;//numero di itemset
	
	double percent;/*percentuale transazion*/

	bool stillElems;//se ci sono ancora insiemi col supporto minimo: per sapere se devo rimanere in ascolto
	unsigned long printTotal;//numero di globally large
	unsigned long unionTotal;//numero delle unioni

	/*array degli ip*/
	char **ip_addresses;
	/*quanto host siamo*/
	int hosts_num;
	/*per stampare i messaggi di output*/
	char outputString[1000];
	/*calcola il numero di transazioni locali e totali*/
	void calculateTrans();
	/*crea un array di numeri rappresentante l'insieme*/
	int* createElem(int *itemset, int length);
	/*analoga alla pruneCandidates non modificata*/
	int pruneCandidates2(int level);
	/*analoga alla pruneCandidates non modificata*/
	int pruneCandidates2(int level, set<Item> *items, int depth, int *itemset);
	/*END NEW*/

protected:
	
	int generateCandidates(int level);
	int generateCandidates(int level, set<Item> *items, int depth, int *current);
	bool checkSubsets(int sl, int *iset, set<Item> *items, int spos, int depth);

	int countCandidates(int level);

	// test k-subsets of transaction that are candidate itemsets
	int processTransaction(int level, Transaction *t, set<Item> *items, int spos, int depth);
	// test all candidates whether contained in transaction
	int processTransaction2(int level, Transaction *t, set<Item> *items, int spos, int depth);

	void ReOrder();

	int pruneNodes(int level);
	int pruneNodes(int level, set<Item> *items, int depth);

	int pruneCandidates(int level);
	int pruneCandidates(int level, set<Item> *items, int depth, int *itemset);

	void printSet(const Item& item, int *itemset, int length);


private:

	Item *trie;

	ofstream setsout;

	Data *data;
	int *remap;
	set<Element> *relist;

	bool verbose;

	int countType;
};
